 /*
  * pmap_check - additional portmap security.
  * 
  * Always reject non-local requests to update the portmapper tables.
  * 
  * Refuse to forward mount requests to the nfs mount daemon. Otherwise, the
  * requests would appear to come from the local system, and nfs export
  * restrictions could be bypassed.
  * 
  * Refuse to forward requests to the nfsd process.
  * 
  * Refuse to forward requests to NIS (YP) daemons; The only exception is the
  * YPPROC_DOMAIN_NONACK broadcast rpc call that is used to establish initial
  * contact with the NIS server.
  * 
  * Always allocate an unprivileged port when forwarding a request.
  * 
  * If compiled with -DCHECK_PORT, require that requests to register or
  * unregister a privileged port come from a privileged port. This makes it
  * more difficult to replace a critical service by a trojan. Also, require
  * that requests to set/unset the NFSD port come form a privileged port.
  * 
  * If compiled with -DHOSTS_ACCESS, reject requests from hosts that are not
  * authorized by the /etc/hosts.{allow,deny} files. The local system is
  * always treated as an authorized host. The access control tables are never
  * consulted for requests from the local system, and are always consulted
  * for requests from other hosts. Access control is based on IP addresses
  * only; attempts to map an address to a host name might cause the
  * portmapper to hang.
  * 
  * Author: Wietse Venema (wietse@wzv.win.tue.nl), dept. of Mathematics and
  * Computing Science, Eindhoven University of Technology, The Netherlands.
  */

#ifndef lint
static char sccsid[] = "@(#) pmap_check.c 1.8 96/07/07 10:49:10";
#endif

#include <sys/signal.h>
#include <syslog.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <smallrpc/rpc.h>
#include <smallrpc/pmap_prot.h>

extern char *inet_ntoa();

#include "pmap_check.h"

/* Explicit #defines in case the include files are not available. */

#define NFSPROG			100003
#define MOUNTPROG		100005
#define	YPXPROG			100069
#define YPPROG         		100004
#define YPPROC_DOMAIN_NONACK	2
#define MOUNTPROC_MNT		1
#define NFS_PORT		2049

static void logit();
static void toggle_verboselog();
int     verboselog = 0;
int     allow_severity = LOG_INFO;
int     deny_severity = LOG_WARNING;

/* A handful of macros for "readability". */

#define	good_client(a) hosts_ctl("portmap", "", inet_ntoa(a->sin_addr), "")

#define reserved_port(p) (IPPORT_RESERVED/2 < (p) && (p) < IPPORT_RESERVED)

#define unreserved_port(p) (IPPORT_RESERVED <= (p) && (p) != NFS_PORT)

#define	legal_port(a,p) \
  (reserved_port(ntohs((a)->sin_port)) || unreserved_port(p))

#define log_bad_port(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request from unprivileged port")

#define log_bad_host(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request from unauthorized host")

#define log_bad_owner(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request from non-local host")

#define	log_no_forward(addr, proc, prog) \
  logit(deny_severity, addr, proc, prog, ": request not forwarded")

#define log_client(addr, proc, prog) \
  logit(allow_severity, addr, proc, prog, "")

/*
 * Need to initialize logging
 *
 *	 signal(SIGINT, toggle_verboselog);
 */

/*
 * check_default - additional checks for NULL, DUMP, GETPORT and unknown
 */
int
check_default(struct svc_req *rqstp, unsigned int prog)
{
	struct sockaddr_in *addr;

	addr = (struct sockaddr_in *) svc_getcaller(rqstp->rq_xprt);
#ifdef HOSTS_ACCESS
	if (!(from_local(addr) || good_client(addr))) {
		log_bad_host(addr, proc, prog);
		return 0;
	}
#endif

	if (verboselog)
		log_client(addr, rqstp->rq_proc, prog);
	return 1;
}

/*
 * check_privileged_port - additional checks for privileged-port updates
 */
int
check_privileged_port(struct svc_req *rqstp, unsigned int prog, unsigned int port)
{
	struct sockaddr_in *addr;

	addr = (struct sockaddr_in *) svc_getcaller(rqstp->rq_xprt);
#ifdef CHECK_PORT
	if (!legal_port(addr, port)) {
		log_bad_port(addr, rqstp->rq_proc, prog);
		return 0;
	}
#endif
	return 1;
}

/*
 * check_setunset - additional checks for update requests
 */
static inline int
check_local_req(SVCXPRT *xprt)
{
	struct sockaddr_in	*sin;

#ifdef LOOPBACK_SETUNSET
	sin = (struct sockaddr_in *) svc_getsvcaddr(xprt);
	return sin->sin_addr.s_addr == htonl(INADDR_LOOPBACK);
#else
	sin = (struct sockaddr_in *) svc_getcaller(xprt);
	return from_local(sin);
#endif
}

int
check_setunset(struct svc_req *rqstp, unsigned int prog, unsigned int port)
{
	struct sockaddr_in *addr;
	unsigned int	proc;
	
	addr = (struct sockaddr_in *) svc_getcaller(rqstp->rq_xprt);
	proc = rqstp->rq_proc;

	if (!check_local_req(rqstp->rq_xprt)) {
#ifdef HOSTS_ACCESS
		/* because of side effects */
		good_client(addr);
#endif
		log_bad_owner(addr, proc, prog);
		return 0;
	}
	if (port && !check_privileged_port(rqstp, prog, port))
		return 0;
	if (verboselog)
		log_client(addr, proc, prog);
	return 1;
}

/*
 * check_callit - additional checks for forwarded requests
 */
int
check_callit(struct svc_req *rqstp, unsigned int prog, unsigned int aproc)
{
	struct sockaddr_in *addr;
	unsigned int	proc;
	
	addr = (struct sockaddr_in *) svc_getcaller(rqstp->rq_xprt);
	proc = rqstp->rq_proc;

#ifdef HOSTS_ACCESS
	if (!(from_local(addr) || good_client(addr))) {
		log_bad_host(addr, proc, prog);
		return 0;
	}
#endif
	if (prog == PMAPPROG || prog == NFSPROG || prog == YPXPROG
	 || (prog == MOUNTPROG && aproc == MOUNTPROC_MNT)
	 || (prog == YPPROG && aproc != YPPROC_DOMAIN_NONACK)) {
		log_no_forward(addr, proc, prog);
		return 0;
	}
	if (verboselog)
		log_client(addr, proc, prog);
	return 1;
}

/*
 * toggle_verboselog - toggle verbose logging flag
 */
static void
toggle_verboselog(int sig)
{
	signal(sig, toggle_verboselog);
	verboselog = !verboselog;
}

/*
 * logit - report events of interest via the syslog daemon
 *
 * The previous implementation would fork and do a getrpcbynumber.
 * I think this is little gain in relation to the potential for a
 * DoS (think NIS traffic).
 *
 * Instead we now snarf the contents of /etc/rpc at startup time
 * and use that. We lose entries in the NIS rpc map, but that's
 * probably not a huge loss.
 */
static struct proc_map {
	unsigned int	code;
	const char	*proc;
} procmap[] = {
      { PMAPPROC_CALLIT, "callit" },
      { PMAPPROC_DUMP, "dump" },
      { PMAPPROC_GETPORT, "getport" },
      { PMAPPROC_NULL, "null" },
      { PMAPPROC_SET, "set" },
      { PMAPPROC_UNSET, "unset" },

      { 0, 0 }
};

static void
logit(int severity, struct sockaddr_in *addr, 
		unsigned int procnum, unsigned int prognum,
		const char *text)
{
	char		procbuf[32], progbuf[32];
	const char	*procname, *progname;
	struct proc_map *procp;

	/*
	 * Fork off a process or the portmap daemon might hang while
	 * getrpcbynumber() or syslog() does its thing.
	 */

	/* Try to map program number to name. */
	if (prognum == 0) {
		progname = "";
	} else
	if ((progname = rpcname_find(prognum)) == NULL) {
		sprintf(progbuf, "%u", prognum);
		progname = progbuf;
	}

	/* Try to map procedure number to name. */
	for (procp = procmap; procp->proc && procp->code != procnum; procp++)
		/* void */ ;
	if ((procname = procp->proc) == 0) {
		sprintf(procbuf, "%u", procnum);
		procname = procbuf;
	}

	/* Write syslog record. */
	syslog(severity, "connect from %s to %s(%s)%s",
		inet_ntoa(addr->sin_addr), procname, progname, text);
}
